import os
import sys

# These scripts need to continue using Python 2 rather than 3, because
# make_msvc_package.py puts the current Python interpreter on the PATH for the
# sake of gyp, and gyp doesn't work with Python 3 yet.
# https://bugs.chromium.org/p/gyp/issues/detail?id=36
if os.name != "nt":
    sys.exit("Error: ship scripts require native Python 2.7. (wrong os.name)")
if sys.version_info[0:2] != (2,7):
    sys.exit("Error: ship scripts require native Python 2.7. (wrong version)")

import glob
import hashlib
import shutil
import subprocess
from distutils.spawn import find_executable

topDir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

with open(topDir + "/VERSION.txt", "rt") as f:
    winptyVersion = f.read().strip()

def rmrf(patterns):
    for pattern in patterns:
        for path in glob.glob(pattern):
            if os.path.isdir(path) and not os.path.islink(path):
                print("+ rm -r " + path)
                sys.stdout.flush()
                shutil.rmtree(path)
            elif os.path.isfile(path):
                print("+ rm " + path)
                sys.stdout.flush()
                os.remove(path)

def mkdir(path):
    if not os.path.isdir(path):
        os.makedirs(path)

def requireExe(name, guesses):
    if find_executable(name) is None:
        for guess in guesses:
            if os.path.exists(guess):
                newDir = os.path.dirname(guess)
                print("Adding " + newDir + " to Path to provide " + name)
                os.environ["Path"] = newDir + ";" + os.environ["Path"]
    ret = find_executable(name)
    if ret is None:
        sys.exit("Error: required EXE is missing from Path: " + name)
    return ret

class ModifyEnv:
    def __init__(self, **kwargs):
        self._changes = dict(kwargs)
        self._original = dict()

    def __enter__(self):
        for var, val in self._changes.items():
            self._original[var] = os.environ[var]
            os.environ[var] = val

    def __exit__(self, type, value, traceback):
        for var, val in self._original.items():
            os.environ[var] = val

def sha256(path):
    with open(path, "rb") as fp:
        return hashlib.sha256(fp.read()).hexdigest()

def checkSha256(path, expected):
    actual = sha256(path)
    if actual != expected:
        sys.exit("error: sha256 hash mismatch on {}: expected {}, found {}".format(
            path, expected, actual))

requireExe("git.exe", [
    "C:\\Program Files\\Git\\cmd\\git.exe",
    "C:\\Program Files (x86)\\Git\\cmd\\git.exe"
])

commitHash = subprocess.check_output(["git.exe", "rev-parse", "HEAD"]).strip()
defaultPathEnviron = "C:\\Windows\\System32;C:\\Windows"

ZIP_TOOL = requireExe("7z.exe", [
    "C:\\Program Files\\7-Zip\\7z.exe",
    "C:\\Program Files (x86)\\7-Zip\\7z.exe",
])

requireExe("curl.exe", [])



# Copyright (C) 2016 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

source("../../shared/qtcreator.py")

def moveDownToNextNonEmptyLine(editor):
    currentLine = "" # there's no do-while in python - so use empty line which fails
    while not currentLine:
        if waitFor("object.exists(':Utils::FakeToolTip')", 100):
            type(editor, "<Esc>")   # close possibly shown completion tooltip so pressing
        type(editor, "<Down>")      # down scrolls the line, not completion alternatives
        currentLine = str(lineUnderCursor(editor)).strip()
    return currentLine

def performAutoCompletionTest(editor, lineToStartRegEx, linePrefix, testFunc, *funcArgs):
    if platform.system() == "Darwin":
        bol = "<Ctrl+Left>"
        eol = "<Ctrl+Right>"
        autoComp = "<Meta+Space>"
    else:
        bol = "<Home>"
        eol = "<End>"
        autoComp = "<Ctrl+Space>"

    if not placeCursorToLine(editor, lineToStartRegEx, True):
        return
    type(editor, bol)
    # place cursor onto the first statement to be tested
    while not str(lineUnderCursor(editor)).strip().startswith(linePrefix):
        type(editor, "<Down>")
    currentLine = str(lineUnderCursor(editor)).strip()
    while currentLine.startswith(linePrefix):
        type(editor, eol)
        type(editor, "<Ctrl+/>")      # uncomment current line
        snooze(1)
        type(editor, autoComp)        # invoke auto-completion
        testFunc(currentLine, *funcArgs)
        type(editor, "<Ctrl+/>")      # comment current line again
        type(editor, bol)
        currentLine = moveDownToNextNonEmptyLine(editor)

def checkIncludeCompletion(editor, isClangCodeModel):
    test.log("Check auto-completion of include statements.")
    # define special handlings
    noProposal = []
    specialHandling = {"ios":"iostream", "cstd":"cstdio"}
    if isClangCodeModel:
        specialHandling["QDe"] = "QDebug"
        for i in specialHandling.keys():
            specialHandling[i] = " %s>" % specialHandling[i]
    else:
        noProposal += ["detail/hea"]

    # define test function to perform the _real_ auto completion test on the current line
    def testIncl(currentLine, *args):
        noProposal, specialHandling = args
        inclSnippet = currentLine.split("//#include")[-1].strip().strip('<"')
        propShown = waitFor("object.exists(':popupFrame_TextEditor::GenericProposalWidget')", 2500)
        test.compare(not propShown, inclSnippet in noProposal,
                     "Proposal widget is (not) shown as expected (%s)" % inclSnippet)
        if propShown:
            proposalListView = waitForObject(':popupFrame_Proposal_QListView')
            if inclSnippet in specialHandling:
                doubleClickItem(':popupFrame_Proposal_QListView', specialHandling[inclSnippet],
                                5, 5, 0, Qt.LeftButton)
            else:
                type(proposalListView, "<Return>")
        changedLine = str(lineUnderCursor(editor)).strip()
        test.verify(changedLine[-1] in '>"/',
                    "'%s' has been completed to '%s'" % (currentLine.lstrip("/"), changedLine))

    performAutoCompletionTest(editor, ".*Complete includes.*", "//#include",
                              testIncl, noProposal, specialHandling)

def checkSymbolCompletion(editor, isClangCodeModel):
    test.log("Check auto-completion of symbols.")
    # define special handlings
    expectedSuggestion = {"in":["internal", "int", "intmax_t"],
                          "Dum":["Dummy", "dummy"], "Dummy::O":["ONE","one"],
                          "dummy.":["one", "ONE", "PI", "v1", "v2", "v3"],
                          "dummy.o":["one", "ONE"], "Dummy::In":["Internal", "INT"],
                          "Dummy::Internal::":["DOUBLE", "one"]
                          }
    missing = ["Dummy::s", "Dummy::P", "dummy.b", "dummy.bla(", "internal.o", "freefunc2"]
    expectedResults = {"Dummy::s":"Dummy::sfunc()",
                       "Dummy::P":"Dummy::PI", "dummy.b":"dummy.bla(", "dummy.bla(":"dummy.bla(",
                       "internal.o":"internal.one", "freefunc2":"freefunc2(",
                       "using namespace st":"using namespace std", "afun":"afunc()"}
    if isClangCodeModel:
        missing = ["dummy.bla(", "freefunc2("]
        expectedSuggestion["internal.o"] = ["one"]
        if platform.system() in ('Microsoft', 'Windows'):
            expectedSuggestion["using namespace st"] = ["std", "stdext"]
        else:
            expectedSuggestion["using namespace st"] = ["std", "struct", "struct template"]
    else:
        expectedSuggestion["using namespace st"] = ["std", "st"]
    # define test function to perform the _real_ auto completion test on the current line
    def testSymb(currentLine, *args):
        missing, expectedSug, expectedRes = args
        symbol = currentLine.lstrip("/").strip()
        timeout = 2500
        propShown = waitFor("object.exists(':popupFrame_TextEditor::GenericProposalWidget')", timeout)
        test.compare(not propShown, symbol in missing,
                     "Proposal widget is (not) shown as expected (%s)" % symbol)
        found = []
        if propShown:
            proposalListView = waitForObject(':popupFrame_Proposal_QListView')
            found = [i.strip() for i in dumpItems(proposalListView.model())]
            diffShownExp = set(expectedSug.get(symbol, [])) - set(found)
            if not test.verify(len(diffShownExp) == 0,
                               "Verify if all expected suggestions could be found"):
                test.log("Expected but not found suggestions: %s" % diffShownExp,
                         "%s | %s" % (expectedSug[symbol], str(found)))
            # select first item of the expected suggestion list
            suggestionToClick = expectedSug.get(symbol, found)[0]
            if isClangCodeModel:
                suggestionToClick = " " + suggestionToClick
            doubleClickItem(':popupFrame_Proposal_QListView', suggestionToClick,
                            5, 5, 0, Qt.LeftButton)

        try:
            multiSuggestionToolTip = waitForObject(
                "{leftWidget={text~='[0-9]+ of [0-9]+' type='QLabel'} type='QToolButton'}", 1000)
            if multiSuggestionToolTip is not None:
                test.log("Closing tooltip containing overloaded function completion.")
                type(editor, "<Esc>")
        except LookupError: # no proposal or tool tip for unambiguous stuff - direct completion
            pass

        changedLine = str(lineUnderCursor(editor)).strip()
        if symbol in expectedRes:
            exp = expectedRes[symbol]
        else:
            exp = (symbol[:max(symbol.rfind(":"), symbol.rfind(".")) + 1]
                   + expectedSug.get(symbol, found)[0]).strip()
        test.compare(changedLine, exp, "Verify completion matches.")

    performAutoCompletionTest(editor, ".*Complete symbols.*", "//",
                              testSymb, missing, expectedSuggestion, expectedResults)

def main():
    examplePath = os.path.join(srcPath, "creator", "tests", "manual", "cplusplus-tools")
    if not neededFilePresent(os.path.join(examplePath, "cplusplus-tools.pro")):
        return
    templateDir = prepareTemplate(examplePath)
    examplePath = os.path.join(templateDir, "cplusplus-tools.pro")
    for useClang in [False, True]:
        with TestSection(getCodeModelString(useClang)):
            if not startCreatorVerifyingClang(useClang):
                continue
            openQmakeProject(examplePath, [Targets.DESKTOP_5_14_1_DEFAULT])
            checkCodeModelSettings(useClang)
            if not openDocument("cplusplus-tools.Sources.main\\.cpp"):
                earlyExit("Failed to open main.cpp.")
                return
            editor = getEditorForFileSuffix("main.cpp")
            if editor:
                if useClang:
                    test.log("Wait for parsing to finish...")
                    progressBarWait(15000)
                    test.log("Parsing done.")
                checkIncludeCompletion(editor, useClang)
                checkSymbolCompletion(editor, useClang)
                invokeMenuItem('File', 'Revert "main.cpp" to Saved')
                clickButton(waitForObject(":Revert to Saved.Proceed_QPushButton"))
            invokeMenuItem("File", "Exit")
            waitForCleanShutdown()



# Copyright (C) 2016 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

def toggleIssuesFilter(filterName, checked):
    try:
        toggleFilter = waitForObject("{type='QToolButton' toolTip='Filter by categories' "
                                     "window=':Qt Creator_Core::Internal::MainWindow'}")
        clickButton(toggleFilter)
        filterMenu = waitForObject("{type='QMenu' unnamed='1' visible='1'}")
        waitFor("filterMenu.visible", 1000)

        filterCategory = waitForObjectItem(filterMenu, filterName, 1000)
        waitFor("filterCategory.visible", 2000)
        if filterCategory.checked == checked:
            test.log("Filter '%s' has already check state %s - not toggling."
                     % (filterName, checked))
            clickButton(toggleFilter) # close the menu again
        else:
            activateItem(filterCategory)
            test.log("Filter '%s' check state changed to %s." % (filterName, checked))
    except:
        t,v = sys.exc_info()[:2]
        test.warning("Exception while toggling filter '%s'" % filterName,
                     "%s: %s" % (t.__name__, str(v)))


def getBuildIssues(ignoreCodeModel=True):
    # Heuristically determine whether the ClandCodeModel is loaded.
    # The current implementation is inaccurate:
    # The value may be "True" although the CCM was not loaded due to existing settings or
    # insufficient memory. This would result in a slightly longer execution and false positive
    # warnings. Since neither would cause an actual damage and a precise handling would require
    # a bigger refactoring, this seems acceptable.
    clangLoaded = " -noload ClangCodeModel" not in currentApplicationContext().commandLine

    ensureChecked(":Qt Creator_Issues_Core::Internal::OutputPaneToggleButton" , silent=True)
    model = waitForObject(":Qt Creator.Issues_QListView").model()
    if ignoreCodeModel and clangLoaded:
        # filter out possible code model issues present inside the Issues pane
        toggleIssuesFilter("Clang Code Model", False)
    result = dumpBuildIssues(model)
    if ignoreCodeModel and clangLoaded:
        # reset the filter
        toggleIssuesFilter("Clang Code Model", True)
    return result

# this method checks the last build (if there's one) and logs the number of errors, warnings and
# lines within the Issues output
# param expectedToFail can be used to tell this function if the build was expected to fail or not
def checkLastBuild(expectedToFail=False):
    buildIssues = getBuildIssues()
    types = [i[1] for i in buildIssues]
    errors = types.count("1")
    warnings = types.count("2")
    gotErrors = errors != 0
    test.verify(not (gotErrors ^ expectedToFail), "Errors: %s | Warnings: %s" % (errors, warnings))
    return not gotErrors

# helper function to check the compilation when build wasn't successful
def checkCompile(expectToFail=False):
    ensureChecked(":Qt Creator_CompileOutput_Core::Internal::OutputPaneToggleButton")
    output = waitForObject(":Qt Creator.Compile Output_Core::OutputWindow")
    waitFor("len(str(output.plainText))>0",5000)
    if compileSucceeded(output.plainText):
        if os.getenv("SYSTEST_DEBUG") == "1":
            test.log("Compile Output:\n%s" % str(output.plainText))
        if expectToFail:
            test.fail("Compile successful - but was expected to fail.")
        else:
            test.passes("Compile successful")
        return True
    else:
        if expectToFail:
            test.passes("Expected fail - Compile output:\n%s" % str(output.plainText))
        else:
            test.fail("Compile Output:\n%s" % str(output.plainText))
        return False

def compileSucceeded(compileOutput):
    return None != re.match(".*finished successfully\.\n\d\d:\d\d:\d\d: Elapsed time: "
                            "(\d:)?\d{2}:\d\d\.$", str(compileOutput), re.S)

def waitForCompile(timeout=60000):
    progressBarWait(10000) # avoids switching to Issues pane after checking Compile Output
    ensureChecked(":Qt Creator_CompileOutput_Core::Internal::OutputPaneToggleButton")
    output = waitForObject(":Qt Creator.Compile Output_Core::OutputWindow")
    if not waitFor("re.match('.*Elapsed time: (\d:)?\d{2}:\d\d\.$', str(output.plainText), re.S)", timeout):
        test.warning("Waiting for compile timed out after %d s." % (timeout / 1000))

def dumpBuildIssues(listModel):
    issueDump = []
    for index in dumpIndices(listModel):
        issueDump.extend([[str(index.data(role).toString()) for role
                           in range(Qt.UserRole, Qt.UserRole + 2)]])
    return issueDump


# returns a list of pairs each containing the ID of a kit (see class Targets)
# and the name of the matching build configuration
# param filter is a regular expression to filter the configuration by their name
def iterateBuildConfigs(filter = ""):
    switchViewTo(ViewConstants.PROJECTS)
    configs = []
    for currentKit in iterateConfiguredKits():
        switchToBuildOrRunSettingsFor(currentKit, ProjectSettings.BUILD)
        model = waitForObject(":scrollArea.Edit build configuration:_QComboBox").model()
        prog = re.compile(filter)
        # for each row in the model, write its data to a list
        configNames = dumpItems(model)
        # pick only those configuration names which pass the filter
        configs += zip([currentKit] * len(configNames),
                       [config for config in configNames if prog.match(config)])
    switchViewTo(ViewConstants.EDIT)
    return configs

# selects a build configuration for building the current project
# param wantedKit specifies the ID of the kit to select (see class Targets)
# param configName is the name of the configuration that should be selected
# param afterSwitchTo the ViewConstant of the mode to switch to after selecting or None
def selectBuildConfig(wantedKit, configName, afterSwitchTo=ViewConstants.EDIT):
    switchViewTo(ViewConstants.PROJECTS)
    if any((switchToBuildOrRunSettingsFor(wantedKit, ProjectSettings.BUILD),
            selectFromCombo(":scrollArea.Edit build configuration:_QComboBox", configName))):
        progressBarWait(30000)
    if afterSwitchTo:
        if ViewConstants.FIRST_AVAILABLE <= afterSwitchTo <= ViewConstants.LAST_AVAILABLE:
            switchViewTo(afterSwitchTo)
        else:
            test.warning("Don't know where you trying to switch to (%s)" % afterSwitchTo)

# This will not trigger a rebuild. If needed, caller has to do this.
def verifyBuildConfig(currentTarget, configName, shouldBeDebug=False, enableShadowBuild=False,
                      enableQmlDebug=False, buildSystem=None):
    selectBuildConfig(currentTarget, configName, None)
    ensureChecked(waitForObject(":scrollArea.Details_Utils::DetailsButton"))

    if buildSystem == "qmake":
        ensureChecked("{leftWidget={text='Shadow build:' type='QLabel' unnamed='1' visible='1' "
                                    "window=':Qt Creator_Core::Internal::MainWindow'} "
                      "type='QCheckBox' unnamed='1' visible='1' "
                      "window=':Qt Creator_Core::Internal::MainWindow'}", enableShadowBuild)

    buildCfCombo = waitForObject(":scrollArea.Edit build configuration:_QComboBox")
    buildType = "Debug" if shouldBeDebug else "Release"
    test.compare(buildCfCombo.currentText, buildType,
                 "Verifying whether it's a %s build" % buildType)
    qmlDebuggingCombo = waitForObject(':Qt Creator.QML debugging and profiling:_QComboBox')
    if (selectFromCombo(qmlDebuggingCombo, "Enable" if enableQmlDebug else "Disable")
        and buildSystem == "qmake"):
        # Don't rebuild now
        clickButton(waitForObject(":QML Debugging.No_QPushButton", 5000))
        # Wait for parsing to finish
        progressBarWait(14000)
    clickButton(waitForObject(":scrollArea.Details_Utils::DetailsButton"))
    switchViewTo(ViewConstants.EDIT)

# verify if building and running of project was successful
def verifyBuildAndRun(expectCompileToFail=False):
    # check compile output if build successful
    checkCompile(expectCompileToFail)
    # check application output log
    appOutput = logApplicationOutput()
    if appOutput:
        strOutput = str(appOutput)
        test.verify((re.search(".* exited with code \d+", strOutput) or
                     re.search(".* crashed\.", strOutput) or
                     re.search(".* was ended forcefully\.", strOutput) or
                     re.search(".* finished successfully\.", strOutput) or
                     re.search(".* was canceled after \d+ ms\.", strOutput)) and
                    re.search('[Ss]tarting.*', strOutput),
                    "Verifying if built app started and closed successfully.")

# run project for debug and release
def runVerify(expectCompileToFailFor=None):
    availableConfigs = iterateBuildConfigs()
    if not availableConfigs:
        test.fatal("Haven't found build configurations, quitting")
        saveAndExit()
    for kit, config in availableConfigs:
        selectBuildConfig(kit, config)
        expectCompileToFail = False
        if expectCompileToFailFor is not None and kit in expectCompileToFailFor:
            expectCompileToFail = True
        test.log("Using build config '%s'" % config)
        if runAndCloseApp(expectCompileToFail) == None:
            checkCompile(expectCompileToFail)
            continue
        verifyBuildAndRun(expectCompileToFail)
        mouseClick(waitForObject(":*Qt Creator.Clear_QToolButton"))
